#!/usr/bin/env python

from mpi4py import MPI
import sys
import numpy as np
import time
from matrix import *

comm = MPI.COMM_WORLD
worldSize = comm.Get_size()
rank = comm.Get_rank()
processorName = MPI.Get_processor_name()

TaskMaster = worldSize - 1
# name = sys.argv[1]
n = TaskMaster
x, y, z = map(lambda x: int(x), [sys.argv[1], sys.argv[2], sys.argv[3]])
D = map(lambda x: int(x), [sys.argv[4], sys.argv[5], sys.argv[6]])
ENC = sys.argv[7]

import logging

if '-v' in sys.argv:
    logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.DEBUG)
else:
    logging.basicConfig(format='%(levelname)s:%(message)s', level=logging.INFO)

if rank >= n and rank < TaskMaster:
    print("invalid rank")
    exit(1)

if ENC == "EP" or ENC == "XEP" or ENC == "2EP":
    k = x * y * z + z - 1
    row = k
elif ENC == "DEP":
    k = 3 * x * y * z / 4 + z / 2 - 1
    row = k
else:
    print("invalid ENC")
    exit(1)


logging.debug("[%d] Process %s started." % (rank, processorName))

def worker(A1, A2, X1, X2, D, TaskMaster):
    if (ENC == "EP"):
        Y = np.dot(A1, X1)
        Y.reshape(1, D[0] * D[2])
        request = comm.Ssend(Y, dest = TaskMaster, tag = 0)
    elif (ENC == "XEP"):
        Y = np.dot(A1, X1)
        Y.reshape(1, D[0] * D[2])
        request = comm.Ssend(Y, dest = TaskMaster, tag = 0)
        Y = np.dot(A2, X2)
        Y.reshape(1, D[0] * D[2])
        request = comm.Ssend(Y, dest = TaskMaster, tag = 0)
    elif ENC == "2EP":
        Y = np.zeros(D[0] * D[2] * 2)
        Y[0 : (D[0] * D[2])] = np.dot(A1, X1).reshape(1, D[0] * D[2])
        Y[(D[0] * D[2]) : (2 * D[0] * D[2])] = np.dot(A2, X2).reshape(1, D[0] * D[2])
        request = comm.Ssend(Y, dest = TaskMaster, tag = 0)
    elif ENC == "DEP":
        Y = np.dot(A1, X1) + np.dot(A2, X2)
        Y.reshape(D[0], D[2])
        request = comm.Ssend(Y, dest = TaskMaster, tag = 0)

def master(n, k, D):
    logging.info("[%d] k = %d" % (rank, k))
    logging.info("[%d] origin = %d" % (rank, row))

    t_start = MPI.Wtime()
    helper = []
    index = []
    M = np.zeros((k, D[0] * D[2]))
    num = 0
    parity = 0

    def print_time():
        time = MPI.Wtime()
        print time - t_start

    Me = np.zeros(2 * D[0] * D[2])
    while(len(index) < k):
        st = MPI.Status()
        comm.Probe(status=st)
        i = st.source
        if (ENC == "EP" or ENC == "DEP" or ENC == "XEP"):
            comm.Recv([M[num], MPI.FLOAT], source = i, tag = 0)
            num += 1
            if (ENC == "XEP") and i in index:
                index.append(i + k)
            else:
                index.append(i)
        else:
            comm.Recv([Me, MPI.FLOAT], source = i, tag = 0)
            Me.reshape(2, D[0] * D[2])
            M[num] = Me[0]
            num +=1
            index.append(i * 2)
            if (num < k):
                M[num] = Me[1]
                num += 1
                index.append(i * 2 + 1)
        logging.debug("[%d] worker %d received" % (rank, i))
        logging.debug("[%d] index received: [%s]" % (rank, " ".join(map(lambda x: str(x), index))))
        if i >= k:
            parity += 1
    logging.info("[%d] Parity: %d" % (rank, parity))
    t_data = MPI.Wtime()
    logging.debug("[%d] Enough results obtained." % (rank))
    logging.debug("[%d] Time: %5.4fs" % (rank, MPI.Wtime() - t_start))
    # logging.debug("[%d] helper: [%s]" % (rank, " ".join(map(lambda x: str(len(x)), helper))))
    G = Vandermonde(k, index)
    logging.debug("[%d] Time: %5.4fs" % (rank, MPI.Wtime() - t_start))
    decode(M, G, row)
    t_finish = MPI.Wtime()
    t_all = t_finish - t_start
    t_decode = t_finish - t_data
    logging.info("[%d] Process %d finished in %5.4fs, decoding in %5.4fs" %(rank, rank, t_all, t_decode))

if rank != TaskMaster:
    logging.debug("[%d] Running from processor %s, rank %d out of %d processors." % (rank, processorName, rank, worldSize))

    t_start = MPI.Wtime()
    A1 = np.random.rand(D[0], D[1])
    A2 = np.random.rand(D[0], D[1])
    X1 = np.random.rand(D[1], D[2])
    X2 = np.random.rand(D[1], D[2])
    comm.Barrier()
    worker(A1, A2, X1, X2, D, TaskMaster)
    exit(0)

if rank == TaskMaster:
    logging.debug("[%d] Checking response from Workers." % (rank))
    comm.Barrier()
    master(n, k, D)
    comm.Abort(0)
    exit(0)
